"""Simple IPC Message Queue class for Python3, using ctypes and libc.
"""
#Adopted from: https://gist.github.com/unautre/3cf4e9924ee81b762f82
__version__ = '1.0.0 2024-04-21'# Transport is static class
#TODO: Currently ftok file does not matter. Issue: if send size is bigger than recv buffer then queue will be spoiled. Ideally it should contain the max size of the receiver buffer.

"""```````````````````Usage
$ python3 parmhost/ipc.py -l
$ python3
from parmhost.ipc import Transport
msq = Transport()
msq.send(b'hello')
"""
import sys

from ctypes import CDLL, get_errno, create_string_buffer
from ctypes import c_ushort, c_int, c_uint, c_long, c_ulong
from ctypes import c_char_p, c_void_p, POINTER, Structure, cast

# we load the libc.
libc = CDLL("libc.so.6")
# define some consts.
IPC_CREAT = 0o1000
IPC_EXCL  = 0o2000
IPC_NOWAIT= 0o4000
IPC_RMID = 0
IPC_SET = 1
IPC_STAT = 2
IPC_INFO = 3
# typedef some types
key_t = c_int

# Static key for IPC    
ftok = libc.ftok
ftok.argtypes = [c_char_p, c_int]
ftkey = ftok(c_char_p(b"/tmp/ipcbor.ftok"), c_int(65))
if ftkey == -1:
    print("ERR: Could not create IPC Message Key, Please do: 'touch  /tmp/ipcbor.ftok'")
    sys.exit(1);
print(f'IPC ftok: {ftkey}')

class ipc_perm(Structure):
    _fields_ = [    ("__key", key_t),
            ("uid", c_uint),
            ("gid", c_uint),
            ("cuid", c_uint),
            ("cgid", c_uint),
            ("mode", c_ushort),
    ]

class msqid_ds(Structure):
    _fields_ = [    ("msg_perm", ipc_perm),
            ("msg_stime", c_ulong),
            ("msg_rtime", c_ulong),
            ("msg_ctime", c_ulong),
            ("__msg_cbytes", c_ulong),
            ("msg_qnum", c_ulong), # Useful:number of messages in queue
            ("msg_qbytes", c_ulong),
            ("msg_lspid", c_int),
            ("msg_lrpid", c_int)
    ]

class Transport:
    """Static class"""
    msgget = libc.msgget
    msgget.argtypes = [key_t, c_int]
    
    msgsnd = libc.msgsnd
    msgsnd.argtypes = [c_int, c_char_p, c_ulong, c_int]
    
    msgctl = libc.msgctl
    msgctl.argtypes = [c_int, c_int, POINTER(msqid_ds)]
    
    msgrcv = libc.msgrcv
    msgrcv.argtypes = [c_int, c_char_p, c_ulong, c_ulong, c_int]
    
    perror = libc.perror
    perror.argtypes = [c_char_p]
    
    def init(key=ftkey, mode=IPC_CREAT | 0o644, bitrate=None):
        Transport.id_rcv = Transport.msgget(key+1, mode)
        Transport.id_snd = Transport.msgget(key, mode)
        print(f'IPC id_rcv: {Transport.id_rcv}, id_snd: {Transport.id_snd}')

    def send(msg):#, msglen), msgtype=1, flag=0):
        msgtype=1; flag=0
        msglen = len(msg)
        buf = create_string_buffer(msglen+8)
        lbuf = cast(buf, POINTER(c_long))
        lbuf[0] = msgtype
        buf[8:] = msg[:msglen]
        r = Transport.msgsnd(Transport.id_snd, buf, msglen, flag)
        if r != 0:
            raise BufferError('in Transport.send.')
            msq.clear()

    def recv(msgsize=1500):
        msgtype=1; flag=0
        size = msgsize+8
        buf = create_string_buffer(size)
        msglen = Transport.msgrcv(Transport.id_rcv, buf, size, msgtype, flag)
        # do not return type
        #lbuf = cast(buf, POINTER(c_long)) # for msgtype        
        #return lbuf.contents.value, buf[8:msglen+8]
        return buf[8:msglen+8]

    def clear():
        Transport.msgctl(Transport.id_rcv, IPC_RMID, None)
        Transport.msgctl(Transport.id_snd, IPC_RMID, None)
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=__doc__
        ,formatter_class=argparse.ArgumentDefaultsHelpFormatter
        ,epilog=f'ipc {__version__}')
    parser.add_argument('-l','--listener', action='store_true', help=\
    'Print incoming messages')
    parser.add_argument('messages', nargs='*', help='Messages to send to server')
    pargs = parser.parse_args()

    msq = Transport()
    for msg in pargs.messages:
        msq.send(msg)
    
    if pargs.listener:
        print('IPC Listening')
        msglen = 1
        while msglen > 0:
           msg = msq.recv()
           print(f'IPC Received {msglen} bytes:`{msg}`')
           msglen = len(msg)
        msq.clear()
        



"""Access to P2Plant process variables with IPC queue link.
Usage:
$ python3
from p2plantaccess import Access as pa
pa.init()
pa.start()
pa.request(["info", ["*"]])
pa.request(["get", ['run',"version"]])
  or:
pa.request('["get", ["run","version"]]')
pa.request(['set', [('run','stop')]])
"""
__version__ = 'v1.0.1 2025-02-06'#
print(f'p2plantAccess {__version__}')

import time
import threading
from collections import deque
import numpy as np
from json import loads
import cbor2

dtype = {# map of CBOR tags to numpy dtype
    72: "int8",
    64: "uint8",
    69: "uint16",
    77: "int16",
    78: "int32",
    70: "uint32",
}
tag = {v:k for k,v in dtype.items()}# map of dtype to CBOR tag

def printi(msg): print(f'inf_p2pa: {msg}')
def printw(msg): print(f'WAR_p2pa: {msg}')
def printe(msg): print(f'ERR_p2pa: {msg}')
def printv(msg):
    if (Access.DBG): print(f'dbg_p2p: {msg}')

class Access:
    """Static class"""
    DBG = False
    receivedList = []
    commandReply = deque()# OCould keep many deliveries
    subscriptionReply = deque()# Only one delivery could be kept here
    started = False
    receivedMap = {}

    event_subscription_received = threading.Event()
    event_reply_received = threading.Event()
    
    queueAndEvents = {
    'reply': (commandReply, event_reply_received),# event_reply_processed),
    'subscription': (subscriptionReply, event_subscription_received),# event_subscription_processed)
    }

    def init(transport='ipc', bitrate=7372800):
        if transport== 'ipc':
            from .ipc import Transport
        Transport.init(bitrate=bitrate)
        Access.transport = Transport

    def start():
        thread = threading.Thread(target=Access._receiver, daemon=True)
        Access.started = True
        thread.start()

    def stop():
        Access.started = False

    def _receiver():
        printi("Receiver started")
        while Access.started:
            bbuf = Access.transport.recv()
            printv(f'Received a message[{len(bbuf)}]: {bbuf}')
            Access.receivedList = cbor2.loads(bbuf)
            if len(Access.receivedList) > 0\
              and Access.receivedList[0] == 'Subscription':
                printv(f'Subscriptions: {Access.receivedList}')
                if len(Access.subscriptionReply) > 0:
                    continue
                Access.subscriptionReply.append(Access.receivedList[1:])
                Access.event_subscription_received.set()
            else:
                Access.commandReply.append(Access.receivedList)
                Access.event_reply_received.set();
        printi("Receiver finished")

    def send(request):
        """Encode object to CBOR and send it to IPC queue"""
        if not Access.started:
            printw("Receiver is not started")
            return 
        cborobj = cbor2.dumps(request)
        Access.transport.send(cborobj)
        #print('')

    #def array(nparray):
    #    """Convert numpy array to CBORTag. For use in send()."""
    #    atag = tag[str(nparray.dtype)]
    #    return cbor2.CBORTag(atag, nparray.tobytes())

    def recv(what:str = 'reply', blocking='block'):
        """Receive data, the what could be 'reply' or 'Subscription'.
        if blocking=='block', the program will be blocked until arrival of 
        new data"""
        #print(f'recv {what}, {blocking}')
        if not Access.started:
            printw("Receiver is not started")
            return
        try:
            queue, evRcv = Access.queueAndEvents[what]
        except:
            return {'ERR':f'recv supports only {Access.queueAndEvents.keys()}'}
        if blocking.startswith("block"):
            evRcv.wait()
        else:
            if not evRcv.is_set():
                return {}
        l = len(queue)
        printv(f'reply received {what}[{l}]')
        Access.receivedMap = {}
        if l == 0:
            return []
        if l == 1:# Handling last item. The flag could be cleared.
            evRcv.clear()
        qitem = deque(queue.pop())
        printv(f'qitem: {qitem}')
        if (len(qitem)%2) != 0:
            # Received odd number of items: that happens if request was incorrect.
            printe(f'Request was not recognized by server: {qitem[0]}')
            return {}
        try:
            while par := qitem.popleft():
                parmap = qitem.popleft()
                if par in Access.receivedMap:
                   Access.receivedMap[par].update(parmap)
                else:
                    Access.receivedMap[par] = parmap
        except IndexError:
            pass
        #print(f'Access.receivedMap: {Access.receivedMap}')
        return Access.receivedMap

    def decode(parName='*'):
        """Translate reply to standard python objects""" 
        outmap = {}
        inmap = Access.receivedMap
        if parName != '*':
            try:
                inmap = {parName: Access.receivedMap[parName]}
            except:
                return {}
        for par,parmap in inmap.items():
            outmap[par] = {}
            for key,v in parmap.items():
                if isinstance(v, cbor2.CBORTag):
                    v = np.frombuffer(v.value, dtype[v.tag])
                    if key == 'v':
                        shape = parmap.get('shape')
                        v = v.reshape(shape)
                    elif key == 't':
                        v = v[0],v[1]
                outmap[par][key] = v
        return outmap

    def valueAndTimestamp(parName:str):
        """Return (value,timestamps) of a parameter"""
        try:
            parmap = Access.decode(parName)[parName]
            v,t = parmap['v'],parmap['t']
            return (v,t)
        except:
            return []

    def request(obj):
        """Send object or text to IPC queue and receive/decode the reply"""
        try:
            if isinstance(obj, str):
                obj = loads(obj)
            printv(f'Sending obj: {obj}')
            Access.send(obj)
            Access.recv()
            r = Access.decode()
            printv(f'rcv: {r}')
            return r
        except Exception as e:
            printe(f'Sending {obj}: {e}')
            return {}




"""Example of communication with P2Plant simulatedADCs.
"""
__version__ = 'v1.0.1 2025-02-03'#
print(f'Version {__version__}')
import sys, time, os
import threading
import numpy as np
import pprint
from json import loads
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore

from .p2plantaccess import Access as PA

UpArrow = b'\x1b[A' #Up arrow sequence
LastCmd = ''

def printTime(): return time.strftime("%m%d:%H%M%S")
def croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt

def printi(msg): print(f'INF@{printTime()}: {msg}')

def printw(msg):
    msg = croppedText(msg)
    print(f'WRN@{printTime()}: {msg}')

def printe(msg):
    msg = croppedText(msg)
    print(f'ERR@{printTime()}: {msg}')

def _printv(msg, level=0):
    if pargs.verbose is None:
        return
    if len(pargs.verbose) >= level:
        print(msg)
def printv(msg):   _printv(msg, 0)
def printvv(msg):  _printv(msg, 1)

def text_request(text:str):
    try:
        obj = loads(text)
        printv(f'Sending obj: {obj}')
        PA.send(obj)
        PA.recv()
        r = PA.decode()
        printv(f'rcv: {r}')
        return r
    except Exception as e:
        print(f'ERR: sending {text}: {e}')
        return {}

curves = []
lasttime = time.time()
fps = 0
def receive_subscription():
    global lasttime,fps
    ct = time.time()
    printInterval = 10#s
    if ct - lasttime >= printInterval:
        lasttime = ct
        txt = ''
        if pargs.keep_alive:
            # request something to inform server that I am alive
            r = PA.request('["get", ["perf"]]');
            txt += f'Perf: {str(r["perf"]["v"])}.'
        txt += f' Frame rate: {round(fps/printInterval,3)} Hz.'
        if not pargs.graph == '':
            TextItem.setText(txt)
        if not pargs.quiet: 
            printi(txt)
        fps = 0
    r = PA.recv('subscription','nonblocking')
    if len(r) == 0:
        return
    try:
        # data received
        fps += 1
        decoded = PA.decode()
        printv(f'data received: {decoded}')
        try:    data = decoded['adc']['v']
        except: return
        lx = len(data[0])
        #x = np.arange(lx)/adc_srate
        x= np.arange(lx)
        printv(f'Data:[{len(data)}]:\n{data}')
        if pargs.graph == '':
            return
        if len(curves) == 0:
            for idx,row in enumerate(data):
                if pargs.graph is None or str(idx+1) in pargs.graph:
                    curve = pg.PlotCurveItem(pen=(idx,len(data)*1.3))
                    plot.addItem(curve)
                    curves.append(curve)
                    legend.addItem(curve, f'ADC{idx+1}')
            plot.resize(900,600)
        for idx,curvrow in enumerate(zip(curves,data)):
            #print(f'idx {idx} {pargs.graph is None or str(idx+1) in pargs.graph}')
            if pargs.graph is None or str(idx+1) in pargs.graph:
                curve,row = curvrow
                curve.setData(x, row*pargs.yscale)
        
    except KeyboardInterrupt:
        print(' Interrupted')
        sys.exit(1)

def input_thread():
    global LastCmd
    print('Accepted commands: info, get, set.')
    while(1):
        msg = input()
        if msg.encode() == UpArrow:
            msg = LastCmd
        else:
            LastCmd = msg
        print(f'requested: {msg}')
        r = PA.request(msg)
        print(f'replied: {r}')
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#``````````````````Main function``````````````````````````````````````````````
def main():
    import argparse
    global pargs, plot, TextItem, legend
    parser = argparse.ArgumentParser(description=__doc__
    ,formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ,epilog=f'{sys.argv[0]}: {__version__}')
    parser.add_argument('-b','--bitrate', type=int, default=7372800, help=
      'Optional, bitrate')
    parser.add_argument('-B','--black', action='store_true', help=
      'Black background, white foreground for all graphics')
    parser.add_argument('-g', '--graph', default='', nargs='?', help=\
      'Show waveforms on a graph, for example, -g: show all, -g123: only show waveforms 1,2 and 3')
    parser.add_argument('-k', '--keep_alive', action='store_true', help=\
      'Periodically send request to server to keep it alive')
    parser.add_argument('-q', '--quiet', action='store_true', help=\
      'Do not print frame statistics')
    parser.add_argument('-t','--transport', default='ipc', choices=['uart','ipc'], help=
      'Transport interface')
    parser.add_argument('-u', '--yunits', default='', help=\
      'Units for Y-axis')
    parser.add_argument('-v', '--verbose', nargs='*', help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('-y', '--yscale', type=float, default=1., help=\
      'Y scale')
    parser.add_argument('-z', '--zoomin', help=\
      'Zoom the application window by a factor')
    pargs = parser.parse_args()
    #breakpoint()

    PA.init(pargs.transport, pargs.bitrate)
    PA.start()

    r = PA.request('["get", ["version"]]');
    print(f'P2Plant {r["version"]["v"]}')
    PA.request('["set", [["run", "start"]]]')

    #```````````````The P2Plant seems to be alive`````````````````````````````````
    threading.Thread(target=input_thread, daemon=True).start()
    if pargs.graph == '':
        # No plotting.
        while(1):
            receive_subscription()
    else:
        if not pargs.black:
            pg.setConfigOption('background', 'w')
            pg.setConfigOption('foreground', 'k')
        # the --zoom should be handled prior to QtWidgets.QApplication
        try:
            print(f'argv: {sys.argv}')
            idx = sys.argv.index('-z')
            zoom = sys.argv[idx+1]
            print(f'Zoom is set to {zoom}')
            os.environ["QT_SCALE_FACTOR"] = zoom
        except Exception as e:
            #print(f'zoom option is not provided: {e}')
            pass
        app = pg.mkQApp()
        plot = pg.plot()
        plot.setWindowTitle('Simulated ADC')
        #print(f'ADC sampling rate: {adc_srate} Hz')
        adc_srate = 1.
        if adc_srate == .9999:
            printw(f'Fail to read ADC sampling rate. It can be recovered using `get adc_srate` command.')
            plot.setLabel('bottom', 'Sample')
        else:
            plot.setLabel('bottom', 'Time', units='S')
        plot.setLabel('left','', units=pargs.yunits)

        legend = pg.LegendItem((80,60), offset=(70,20))
        legend.setParentItem(plot.graphicsItem())

        #text = pg.TextItem(html='<div style="text-align: center"><span style="color: #FFF;">This is the</span><br><span style="color: #FF0; font-size: 16pt;">PEAK</span></div>', anchor=(-0.3,0.5), angle=45, border='w', fill=(0, 0, 255, 100))
        TextItem = pg.TextItem('Acquisition started', color='darkBlue', anchor=(0,1))#, ensureInBounds=True)
        TextItem.setParentItem(plot.graphicsItem())
        #plot.addItem(TextItem)
        #width = plot.size().width()
        #print(f'width: {width}')
        TextItem.setPos(40, 20)
        
        timer = QtCore.QTimer()
        timer.timeout.connect(receive_subscription)
        timer.start(0)
        app.exec_()

if __name__ == "__main__":
    main()

